"""
    Extension for the python ``click`` module
    to provide a group or command with aliases.
"""

# Credits: https://github.com/click-contrib/click-aliases

import click


class ClickAliasedGroup(click.Group):
    def __init__(self, *args, **kwargs):
        super(ClickAliasedGroup, self).__init__(*args, **kwargs)
        self._commands = {}
        self._aliases = {}

    def command(self, *args, **kwargs):
        aliases = kwargs.pop('aliases', [])
        decorator = super(ClickAliasedGroup, self).command(*args, **kwargs)
        if not aliases:
            return decorator

        def _decorator(f):
            cmd = decorator(f)
            if aliases:
                self._commands[cmd.name] = aliases
                for alias in aliases:
                    self._aliases[alias] = cmd.name
            return cmd

        return _decorator

    def group(self, *args, **kwargs):
        aliases = kwargs.pop('aliases', [])
        decorator = super(ClickAliasedGroup, self).group(*args, **kwargs)
        if not aliases:
            return decorator

        def _decorator(f):
            cmd = decorator(f)
            if aliases:
                self._commands[cmd.name] = aliases
                for alias in aliases:
                    self._aliases[alias] = cmd.name
            return cmd

        return _decorator

    def get_command(self, ctx, cmd_name):
        if cmd_name in self._aliases:
            cmd_name = self._aliases[cmd_name]
        command = super(ClickAliasedGroup, self).get_command(ctx, cmd_name)
        if command:
            return command

    def format_commands(self, ctx, formatter):
        rows = []
        for sub_command in self.list_commands(ctx):
            cmd = self.get_command(ctx, sub_command)
            if cmd is None:
                continue
            if cmd.hidden:
                continue
            if sub_command in self._commands:
                aliases = ','.join(sorted(self._commands[sub_command]))
                sub_command = '{0} ({1})'.format(sub_command, aliases)
            cmd_help = cmd.get_short_help_str() or ''
            rows.append((sub_command, cmd_help))
        if rows:
            with formatter.section('Commands'):
                formatter.write_dl(rows)



import json
import os
from pathlib import Path
from time import time

import appdirs

CACHE_DIR = appdirs.user_cache_dir(
    appname='storyscript-cli', appauthor='asyncy'
)
STORAGE_DIR = appdirs.user_state_dir(
    appname='storyscript-cli', appauthor='asyncy'
)


class Storage:
    def __init__(self, path):
        self.path = path
        self._data = {}

        self._touch()
        self._load()
        self._ensure()

    def remove_file_on_disk(self):
        os.remove(self.path)

    def _touch(self):
        # Make the directory, if it doesn't exist.
        if os.path.dirname(self.path) != '':
            # Can be empty if the path is relative, and in the CWD.
            os.makedirs(os.path.dirname(self.path), exist_ok=True)

        # Touch the file.
        with open(self.path, 'a') as f:
            f.write('')

    def _load(self):
        with open(self.path, 'r') as f:
            try:
                self._data.update(json.load(f))
            except ValueError:
                pass

    def _ensure(self):
        for k in self._data:
            if k.startswith('_'):
                if '_expires' in k:
                    expires_key = f'_{k}_expires'
                    try:
                        if time() > self._data[expires_key]:
                            self.delete(k)
                            self.delete(expires_key)

                    except KeyError:
                        pass

    def _save(self):
        with open(self.path, 'w') as f:
            json.dump(self._data, f)

    def fetch(self, key, default=None):
        return self._data.get(key, default)

    def store(self, key, value, expires=False):
        self._data[key] = value

        # Set expiration, if applicable.
        if expires:
            self._data[f'_{key}_expires'] = time() + expires

        self._save()

    def copy_from(self, path, delete=False):
        try:
            # Ensure that the path exists.
            assert os.path.exists(path)

            # Open and de-serialize the old file.
            with open(path, 'r') as f:
                old_contents = json.load(f)

            # Copy the items over.
            for (k, v) in old_contents.items():
                self.store(k, v)

            # Delete the file, if requested.
            if delete:
                os.remove(path)

        except AssertionError:
            pass

    def delete(self, key):
        del self._data[key]
        self._save()

    def as_dict(self):
        return self._data

    def __contains__(self, *args, **kwargs):
        return self._data.__contains__(*args, **kwargs)

    def __getitem__(self, *args, **kwargs):
        return self._data.__getitem__(*args, **kwargs)


cache: Storage = None
config: Storage = None


def init_storage(config_loc: str = None):
    global cache, config

    cache_file = os.path.join(CACHE_DIR, 'cache.json')
    cache = Storage(path=cache_file)

    if config_loc is None:
        config_loc = os.path.join(STORAGE_DIR, 'config.json')

    config = Storage(path=config_loc)
    config.copy_from(f'{Path.home()}/.storyscript/config.json',
                     delete=True)



# -*- coding: utf-8 -*-
from pytest import mark


@mark.parametrize('get_result', [
    {},
    {'service1': {'a': 'b'}, 'c': 'd', 'service2': {'e': 'f'}}
])
def test_list(runner, patch, init_sample_app_in_cwd, get_result):
    with runner.runner.isolated_filesystem():
        init_sample_app_in_cwd()

        from story import api
        from story.commands.config import list_command

        patch.object(api.Config, 'get', return_value=get_result)

        result = runner.run(list_command, exit_code=0)

    api.Config.get.assert_called_with('my_app')

    if get_result:
        assert """
        Storyscript variables:
c:  d

Service variables:
service1
  a:  b
service2
  e:  f
""".strip() in result.stdout
    else:
        assert 'No configuration set yet' in result.stdout
        assert 'story config set key=value' in result.stdout
        assert 'story config set service.key=value' in result.stdout


@mark.parametrize('message', ['my message', None])
def test_set(runner, patch, init_sample_app_in_cwd, message):
    args = ['a=b', 'service1.c=d', 'd=f', 'service2.g=h', 'service1.i=j']

    if message:
        args.append('-m')
        args.append(message)

    expected_variables = {
        'A': 'b',
        'D': 'f',
        'service1': {
            'C': 'd',
            'I': 'j'
        },
        'service2': {
            'G': 'h',
            'old_service_var': 'old_val'
        },
        'old_ss_var': 'secret***'
    }

    old_vars = {
        'old_ss_var': 'secret***',
        'service2': {
            'old_service_var': 'old_val'
        }
    }

    with runner.runner.isolated_filesystem():
        init_sample_app_in_cwd()

        from story import api
        from story.commands.config import set_command

        patch.object(api.Config, 'get', return_value=old_vars)
        patch.object(api.Config, 'set',
                     return_value={'id': 'super_version_number'})

        result = runner.run(set_command, args=args, exit_code=0)

    api.Config.get.assert_called_with(app='my_app')
    api.Config.set.assert_called_with(config=expected_variables,
                                      app='my_app', message=message)

    assert 'Deployed new release' in result.stdout
    assert 'super_version_number' in result.stdout


def test_set_invalid_pair(runner, patch, init_sample_app_in_cwd):
    args = ['ab', 'a=b']

    with runner.runner.isolated_filesystem():
        init_sample_app_in_cwd()

        from story import api
        from story.commands.config import set_command

        patch.object(api.Config, 'get', return_value={})

        result = runner.run(set_command, args=args, exit_code=1)

    assert 'Config variables must be of the form name=value' in result.stdout
    assert 'Got unexpected pair "ab"' in result.stdout


@mark.parametrize('conf_key_name,get_return_value,stdout_asserts', [
    (
        'service_1',
        {
            'service_1': {
                'a': 'b',
                'c': 'd'
            }
        },
        [
            'A:  b',
            'C:  d'
        ]
    ),
    (
        'service_1.a',
        {
            'service_1': {
                'A': 'b'
            }
        },
        [
            'A:  b'
        ]
    ),
    (
        'service_1',
        {
            'SERVICE_1': {
                'A': 'b'
            }
        },
        [
            'A:  b'
        ]
    ),
    (
        'service_1.a',
        {},
        None
    ),
    (
        'key_1',
        {
            'key_1': 'val_1'
        },
        [
            'KEY_1:  val_1'
        ]
    ),
])
def test_get_for_service(runner, patch, init_sample_app_in_cwd,
                         get_return_value, stdout_asserts, conf_key_name):
    args = [conf_key_name]

    with runner.runner.isolated_filesystem():
        init_sample_app_in_cwd()

        from story import api
        from story.commands.config import get

        patch.object(api.Config, 'get', return_value=get_return_value)

        result = runner.run(get, args=args, exit_code=0)

    api.Config.get.assert_called_with(app='my_app')

    assert 'Fetching config for my_app' in result.stdout

    if stdout_asserts is None:
        assert f'No variable named "{conf_key_name}"'
    else:
        for line in stdout_asserts:
            assert line in result.stdout



# -*- coding: utf-8 -*-
import sys

from story.cli import cli


def test_cli_as_module(spawn_cli_subprocess):
    # We don't need space-indents from docstrings
    output = str(cli.__doc__).replace('  ', ' ').strip()
    exit_code, stdout, stderr = spawn_cli_subprocess()

    assert exit_code == 0
    assert output in stdout
    assert stderr == ''



# -*- coding: utf-8 -*-

import click

import emoji

from .. import cli


@cli.cli.command()
def run():
    """Write Storyscript, interactively."""
    click.echo('This command has not been implemented yet.')
    # cli.user()
    #
    # from storyscript import version
    # click.echo(
    #     click.style('Asyncy', fg='magenta') + ' ' +
    #     cli.version + click.style(' - ', dim=True) +
    #     click.style('Storyscript', fg='cyan') + ' ' +
    #     version
    # )
    #
    # from .repl.Repl import Repl
    # try:
    #     Repl().interact()
    #     click.echo(emoji.emojize(':waving_hand:  Goodbye.'))
    # except KeyboardInterrupt:
    #     click.echo(emoji.emojize(':waving_hand:  Goodbye.'))



# -*- coding: utf-8 -*-
from random import choice, randint


left = (
    'cellular',
    'comsmic',
    'constant',
    'crystal',
    'electronic',
    'expansive',
    'light',
    'liquid',
    'lunar',
    'lunar',
    'magnetic',
    'mercious',
    'nebulous',
    'neon',
    'orbital',
    'phasic',
    'quantum',
    'solar',
    'solid',
    'spaceous',
    'super',
    'superfluid',
    'universal',
)

right = (
    'aluminium',
    'andromeda',
    'antimatter',
    'apollo',
    'argon',
    'arsenic',
    'asteroid',
    'atom',
    'belt',
    'beryllium',
    'bismuth',
    'blackhole',
    'boron',
    'boson',
    'bromine',
    'cadmium',
    'calcium',
    'carbon',
    'charge',
    'chlorine',
    'chromium',
    'cobalt',
    'collider',
    'comet',
    'copper',
    'cosmos',
    'curiosity',
    'deimos',
    'europa',
    'fermion',
    'fluorine',
    'force',
    'galaxy',
    'gallium',
    'gaseous',
    'germanium',
    'gold',
    'hadron',
    'helium',
    'higgs',
    'hydrogen',
    'indium',
    'iodine',
    'ion',
    'iridium',
    'iron',
    'jupiter',
    'kepler',
    'krypton',
    'lead',
    'lepton',
    'lithium',
    'magnesium',
    'mars',
    'mercury',
    'meteor',
    'molecule',
    'molecule',
    'molybdenum',
    'moon',
    'neptune',
    'neutrino',
    'neutron',
    'nickel',
    'niobium',
    'nitrogen',
    'opportunity',
    'oxygen',
    'palladium',
    'phobos',
    'phosphorus',
    'photon',
    'plasma',
    'platinum',
    'polymer',
    'potassium',
    'quark',
    'rhodium',
    'rocket',
    'rover',
    'rubidium',
    'ruthenium',
    'satellite',
    'scandium',
    'selenium',
    'silicon',
    'silver',
    'sirius',
    'sodium',
    'sol',
    'sputnik',
    'star',
    'strontium',
    'sulfur',
    'supernova',
    'technetium',
    'tin',
    'titanium',
    'uranium',
    'uranus',
    'vector',
    'venus',
    'voyager',
    'wormhole',
    'yttrium',
    'zinc',
    'zirconium',
)


def new():
    return f'{choice(left)}-{choice(right)}-{randint(0, 9999):04}'



# -*- coding: utf-8 -*-
import os

import click_completion

from pytest import mark

from story.commands import completion


@mark.parametrize('insensitive', [True, False])
@mark.parametrize('shell', ['bash', 'fish', 'zsh', 'powershell', None])
@mark.parametrize('install', [True, False])
@mark.parametrize('path', ['my_completion_file_path', None])
def test_completion(runner, patch, insensitive, shell, install, path):
    patch.object(click_completion.core, 'get_code',
                 return_value='super_cool_completion_code')

    patch.object(click_completion.core, 'install',
                 return_value=(shell, path))

    args = []
    if shell:
        args.append(shell)

    if path:
        args.append('--path')
        args.append(path)

    if install:
        args.append('--install')

    expected_extra_env = {}

    if insensitive:
        args.append('-i')
        expected_extra_env[
            '_CLICK_COMPLETION_COMMAND_CASE_INSENSITIVE_COMPLETE'] = 'ON'

    with runner.runner.isolated_filesystem():
        result = runner.run(completion.completion, args=args, exit_code=0)

    if install:
        click_completion.core.install.assert_called_with(
            shell=shell, path=path, append=True,
            extra_env=expected_extra_env)

        click_completion.core.get_code.assert_not_called()

        assert f'{shell} completion installed in {path}' in result.output
    else:
        click_completion.core.get_code.assert_called_with(
            shell, extra_env=expected_extra_env)

        click_completion.core.install.assert_not_called()
        assert 'super_cool_completion_code' in result.output


def test_custom_startswith():
    assert completion.custom_startswith('full_string', 'full') is True
    assert completion.custom_startswith('full_string', 'not_full') is False


def test_custom_startswith_insensitive():
    os.environ['_CLICK_COMPLETION_COMMAND_CASE_INSENSITIVE_COMPLETE'] = 'ON'

    assert completion.custom_startswith('full_string', 'full') is True
    assert completion.custom_startswith('full_string', 'FULL') is True
    assert completion.custom_startswith('full_string', 'fULl') is True
    assert completion.custom_startswith('full_string', 'not_full') is False
    assert completion.custom_startswith('full_string', 'NOT_full') is False



# # -*- coding: utf-8 -*-
#
# import json
# import os
# import re
# import sys
#
# import click
#
# import click_spinner
#
# import emoji
#
# from prompt_toolkit import PromptSession
# from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
# from prompt_toolkit.history import FileHistory
# from prompt_toolkit.key_binding import KeyBindings
# from prompt_toolkit.lexers import PygmentsLexer
# from prompt_toolkit.styles import Style
#
# from pygments.lexers import PythonLexer
#
# import requests
#
# import storyscript
# from storyscript import compiler, parser
#
# from .Scope import Scope
# from ... import cli
#
#
# class Repl:
#     _patterns = re.compile(r'((.* as (\w+(, )?)+)|try|catch|finally)$').match
#
#     toolbar = [
#         ('class:toolbar-key', ' Shift+Tab'),
#         ('class:toolbar-text', ' to end scope'),
#         ('class:toolbar-key', ' /data'),
#         ('class:toolbar-text', ' will show variables'),
#         ('class:toolbar-key', ' /save /help /exit'),
#     ]
#
#     # https://python-prompt-toolkit.readthedocs.io/en/latest/
#     def __init__(self):
#         self.session = PromptSession(
#             history=FileHistory(f'{cli.home}/.history')
#         )
#         self.auto_suggest = AutoSuggestFromHistory()
#         self.compile = compiler.Compiler().compile
#         self.parse = parser.Parser().parse
#         self.scope = Scope()
#         self.kb = self.keybinding()
#         self.lexer = PygmentsLexer(PythonLexer)
#         self.requests = requests.Session()
#         self.style = Style.from_dict({
#             'toolbar-text': '#4512ab bg:#aaaaaa',
#             'toolbar-key': '#4512ab bg:#ffffff',
#             'bottom-toolbar': '#4512ab bg:#4512ab',
#         })
#         self.story = []
#         self.block = []
#
#     def prompt_text(self):
#         if len(self.scope) > 1:
#             return '...' + (' ' * (len(self.scope) - 1) * 4)
#         else:
#             return '>>> '
#
#     def should_indent(self, line):
#         return (
#             line.startswith(('if ', 'unless ', 'else if ', 'else ',
#                              'when ', 'while ', 'function ')) or
#             self._patterns(line)
#         )
#
#     def interact(self):
#         while 1:
#             user_input = self.session.prompt(
#                 self.prompt_text(),
#                 lexer=self.lexer,
#                 key_bindings=self.kb,
#                 bottom_toolbar=lambda: self.toolbar,
#                 style=self.style,
#                 auto_suggest=self.auto_suggest
#             )
#
#             if user_input == 1 or user_input == 'exit':
#                 break
#
#             elif user_input == 0:
#                 lines = self.write_block()
#                 self.block = []
#                 self.submit(lines)
#
#             elif user_input[0] == '/':
#               getattr(self, f'command_{user_input[1:]}', self.command_none)()
#
#             elif user_input in self.scope:
#                 # echo out a value
#                 click.echo(self.scope[user_input])
#
#             elif self.should_indent(user_input):
#                 self.write_in_block(user_input)
#
#             elif user_input != '':
#                 lines = self.write(user_input)
#                 if lines and self.block == []:
#                     self.submit(lines)
#
#     def write_block(self):
#         try:
#             return self.compile(self.parse('\n'.join(self.block)))
#
#         except Exception as e:
#             self.handle_exception(e)
#
#     def write_in_block(self, user_input):
#         try:
#             # syntax check this line
#             compiler.Compiler().compile(
#                 self.parse(f'{user_input.strip()}\n    a = 1\n')
#             )
#             # append to story
#             line = f'{self.scope.indent()}{user_input}'
#             self.story.append(line)
#             self.block.append(line)
#             # enter the scope
#             self.scope.add()
#
#         except Exception as e:
#             self.handle_exception(e)
#
#     def write(self, user_input):
#         try:
#             # add block to compiler
#             lines = self.compile(self.parse(user_input))
#             self.story.append(f'{self.scope.indent()}{user_input}')
#             if self.scope:
#                 self.block.append(f'{self.scope.indent()}{user_input}')
#             return lines
#
#         except Exception as e:
#             self.handle_exception(e)
#
#     def handle_exception(self, exception):
#         click.echo(click.style(str(exception), fg='red'))
#
#     def submit(self, lines):
#         with click_spinner.spinner():
#             # TODO assert service exist in Hub
#             #      the hub data should be built locally to quicker requests
#
#             # send to engine
#             res = self.requests.post(
#                 'http://engine.asyncy.net/interact',
#                 data=lines,
#                 timeout=10
#             )
#             res.raise_for_status()
#             data = res.json()
#
#         self.handle_output(data)
#
#     def handle_output(self, data):
#         """
#         Handle output from the Engine
#         """
#         if data.get('output'):
#             click.echo(data['output'])
#
#         elif data.get('error'):
#             click.echo(click.echo(data['error'], fg='red'))
#
#         if data.get('context'):
#             self.scope.update(data['context'])
#
#     def command_exit(self):
#         click.echo(emoji.emojize(':waving_hand:  Goodbye.'))
#         sys.exit(0)
#
#     def command_save(self):
#         path = click.prompt(emoji.emojize('Write to'))
#         self.save(path)
#         sys.exit(0)
#
#     def command_none(self):
#         click.echo('Command not found')
#
#     def command_data(self):
#         click.echo(self.scope.dumps())
#
#     def command_story(self):
#         click.echo('\n'.join(self.story))
#
#     def command_block(self):
#         click.echo('\n'.join(self.block))
#
#     def keybinding(self):
#         kb = KeyBindings()
#
#         @kb.add('s-tab')
#         def kb_stab(event):
#             if len(self.scope) > 1:
#                 self.scope.pop()
#             event.app.exit(0)
#
#         # TODO capture ctr-d events and exit safely
#         @kb.add('c-d')
#         def kb_cb(event):
#             event.app.exit(1)
#
#         return kb
#
#     def save(self, filepath):
#         """
#         Save the stort to a file path
#         """
#         try:
#             os.makedirs(os.path.basename(filepath))
#         except Exception:
#             pass
#
#         if os.path.exists(filepath):
#             assert click.confirm('Override')
#
#         with open(filepath, 'w+') as file:
#             file.write('\n'.join(self.story) + '\n')



# -*- coding: utf-8 -*-

import click

from .. import cli


@cli.cli.command(hidden=True)
def status():
    """Show Storyscript Cloud status."""
    # TODO get asyncy component
    click.echo('Sorry, command not programmed yet.')
